/*--------------------------------------------------------------------------+
|                                                                          |
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.rcp.application.advisors;

import static java.util.stream.Collectors.toList;
import static org.eclipse.core.resources.ResourcesPlugin.getWorkspace;
import static org.eclipse.wb.swt.SWTResourceManager.getColor;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.activateConfiguration;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.deactivateConfiguration;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.getActiveConfigurationsTransitively;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.text.DecimalFormat;
import java.util.List;

import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.ISelectionService;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.menus.WorkbenchWindowControlContribution;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.af3.project.model.IProjectConfigurationElement;
import org.fortiss.af3.project.model.configuration.DevelopmentProcessConfiguration;
import org.fortiss.af3.project.model.development.ConstraintBasedDevelopmentProcess;
import org.fortiss.af3.project.model.development.DummyDevelopmentProcess;
import org.fortiss.af3.project.model.impl.ProjectConfigurationImpl;
import org.fortiss.tooling.base.model.element.ConstraintConfiguration;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.SuccessConstraintInstanceStatus;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;

// TODO (SB, 3): Class comment
/**
 *
 * @author rahman
 * 
 */
class CurrentObjectiveContributionItem extends WorkbenchWindowControlContribution
		implements ISelectionListener, IResourceChangeListener {

	/** The initial objectives. */
	private String[] initialSetOfObjectives = {};

	/** Placeholder string for when no objective is selected. */
	private String initialObjective = "No modeling objective selected";

	/** The Combo holding the time between the simulation steps. */
	private CCombo currentObjectiveCombo;

	/** The control object holding the combo. */
	// private ToolItem toolitem;

	/** Composite instance to hold several controls */
	private Composite composite;

	/** Configuration of the current project selected in the combo box. */
	private ConstraintBasedDevelopmentProcess currentDevelopmentProcess = null;

	/** Constructor. */
	public CurrentObjectiveContributionItem() {
		super("CurrentObjectiveContributionItem");
		// TODO (SB, 12): See
		// org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils.getSelectionService()
		ISelectionService ss =
				PlatformUI.getWorkbench().getActiveWorkbenchWindow().getSelectionService();

		ss.addSelectionListener(this);
		getWorkspace().addResourceChangeListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public boolean isDynamic() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	protected Control createControl(Composite parent) {
		composite = new Composite(parent, SWT.NONE);
		GridLayout layout = new GridLayout(2, false);
		layout.marginHeight = 0;
		composite.setLayout(layout);

		Label label = new Label(composite, SWT.TRANSPARENT);
		label.setText("Current Objective");

		// Objective Combo
		currentObjectiveCombo = new CCombo(composite, SWT.DROP_DOWN);
		currentObjectiveCombo.setItems(initialSetOfObjectives);
		currentObjectiveCombo.setText(initialObjective);
		currentObjectiveCombo.setToolTipText("Set the next modelling objective.");
		currentObjectiveCombo.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				String textInCombo = ((Combo)e.getSource()).getText();
				if(!textInCombo.isEmpty()) {
					updateObjectiveByName(textInCombo);
				}
			}
		});

		parent.setSize(composite.computeSize(SWT.DEFAULT, SWT.DEFAULT, true));
		return composite;
	}

	/** Compute Combo width. */
	@Override
	protected int computeWidth(Control control) {
		return control.computeSize(SWT.DEFAULT, SWT.DEFAULT, true).x;
	}

	/**
	 * Get the pointer to the configuration having a given name.
	 * 
	 * @param constraintName
	 *            The name of the constraint to look for.
	 * @return The real configuration.
	 */
	private ConstraintConfiguration getConstraintConfigByName(String constraintName) {
		ConstraintConfiguration ret = null;
		if(currentDevelopmentProcess != null) {
			EList<ConstraintConfiguration> configs = currentDevelopmentProcess.getConfigurations();
			for(ConstraintConfiguration constraintConfiguration : configs) {
				if(constraintConfiguration.getName().equals(constraintName)) {
					ret = constraintConfiguration;
				}
			}
		}
		return ret;
	}

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		// Updates the current objective if any object is selected on the navigator window.

		EList<ConstraintConfiguration> configs = null;
		if(selection instanceof ITreeSelection) {
			if(((ITreeSelection)selection).getFirstElement() != null) {
				Object objectChosenElement = ((ITreeSelection)selection).getFirstElement();

				if(objectChosenElement instanceof EObject) {
					EObject chosenElement = (EObject)objectChosenElement;

					FileProject fileProject = null;
					if(chosenElement instanceof FileProject) {
						fileProject = (FileProject)chosenElement;
					} else {
						fileProject = getFirstParentWithType(chosenElement, FileProject.class);
						if(fileProject == null) {
							// We are in the context of a Library.
							return;
						}
					}

					EList<IProjectRootElement> rootElems = fileProject.getRootElements();

					// TODO (SB, 14): Test and justify that list is not empty
					// there only exists one root for each project
					ProjectConfigurationImpl config = (ProjectConfigurationImpl)rootElems.get(0);
					EList<IProjectConfigurationElement> configElements =
							config.getConfigurationElements();

					// TODO (SB, 14): Test and justify that list contains at least two elements (or
					// exactly two elements, what-ever makes more sense)
					DevelopmentProcessConfiguration processConfig =
							(DevelopmentProcessConfiguration)configElements.get(1);

					if(!(processConfig.getDevelopment() instanceof DummyDevelopmentProcess)) {
						ConstraintBasedDevelopmentProcess process =
								(ConstraintBasedDevelopmentProcess)processConfig.getDevelopment();

						configs = process.getConfigurations();
						currentDevelopmentProcess = process;

						List<String> configNames =
								configs.stream().map(c -> c.getName()).collect(toList());

						String[] configNamesArray = new String[configNames.size()];
						configNames.toArray(configNamesArray);
						currentObjectiveCombo.setItems(configNamesArray);

						String currObjName =
								currentDevelopmentProcess.getCurrentObjective().getName();
						currentObjectiveCombo.setText(currObjName);
						updateObjectiveByName(currObjName);
					}
				}
			}
		}
	}

	/**
	 * Helper structure to keep the number of instances of a constraint that are satisfied
	 * versus the total number of instances of that constraint.
	 */
	private class ConstraintResultHolder {
		// TODO(SB, 4): Use text from Constructor parameter documentation
		// TODO(SB, do you really need a Float object here?)
		/** successful constraints count */
		private Float successCount;

		// TODO(SB, 4): Use text from Constructor parameter documentation
		// TODO(SB, do you really need a Float object here?)
		/** total constraints count */
		private Float totalCount;

		// TODO: Just document with /** Constructor. */
		/**
		 * Getter and setter for the number of instances of a constraint that are satisfied
		 * versus the total number of instances of that constraint.
		 * 
		 * @param satisfiedConstraintCount
		 *            the count of constraints that are satisfied
		 * @param totalConstraintCount
		 *            the total number of constraints to be considered
		 */
		public ConstraintResultHolder(float satisfiedConstraintCount, float totalConstraintCount) {
			this.successCount = satisfiedConstraintCount;
			this.totalCount = totalConstraintCount;
		}
	}

	/**
	 * Calculate the scores for the current objectives and all objectives it depends on, display it
	 * in the tool tip and set the background color of the current objective accordingly.
	 */
	private void generateScoresAndUpdateUI(ConstraintBasedDevelopmentProcess process) {

		float successPercentage = -1f;
		ConstraintResultHolder resultHolder = generateScoresForConfig(process);
		if(resultHolder.totalCount == 0) {
			currentObjectiveCombo.setBackground(null);
			currentObjectiveCombo.setToolTipText(
					new DecimalFormat("##").format(0 * 100) + "% of the constraints are satisfied");
		} else {
			successPercentage = resultHolder.successCount / resultHolder.totalCount;
			if(successPercentage == 0) {
				RGB color = new RGB(0, .5f, 1);
				currentObjectiveCombo.setBackground(getColor(color));
			} else if(successPercentage < 1f && successPercentage > 0f) {
				RGB color = new RGB(20, .5f, 1);
				currentObjectiveCombo.setBackground(getColor(color));
			} else if(successPercentage == 1f) {
				RGB color = new RGB(130, .5f, 1);
				currentObjectiveCombo.setBackground(getColor(color));
			}
			String tooltiptext = new DecimalFormat("##").format(successPercentage * 100) +
					"% of the constraints are satisfied";
			currentObjectiveCombo.setToolTipText(tooltiptext);
		}
	}

	/**
	 * Updates the current objective to the objective having the name passed as a parameter.
	 * 
	 * @param newObjectiveName
	 *            The name of the objective to update to.
	 */
	public void updateObjectiveByName(String newObjectiveName) {

		ConstraintConfiguration newObjective = getConstraintConfigByName(newObjectiveName);
		ConstraintConfiguration currentObjective = currentDevelopmentProcess.getCurrentObjective();
		final ConstraintConfiguration oldObjective = currentObjective;
		if(newObjective != null) {
			IPersistencyService ps = IPersistencyService.getInstance();
			ITopLevelElement modelContext = ps.getTopLevelElementFor(currentDevelopmentProcess);
			modelContext.runAsNonDirtyingCommand(() -> {
				currentDevelopmentProcess.setCurrentObjective(newObjective);
				List<ConstraintConfiguration> activeConfigs =
						getActiveConfigurationsTransitively(newObjective, null);
				if(!activeConfigs.contains(oldObjective)) {
					IConstraintInstanceContainer constraintInstanceContainer =
							currentDevelopmentProcess.getConstraintInstanceContainer();
					deactivateConfiguration(oldObjective, constraintInstanceContainer,
							newObjective);
				}
				// TODO(SB, 11): Can't you just shift the definition of
				// constraintInstanceContainer in the above if one level up? Or can
				// deactivateConfiguration() influence the result of
				// currentDevelopmentProcess.getConstraintInstanceContainer()?
				IConstraintInstanceContainer constraintInstanceContainer =
						currentDevelopmentProcess.getConstraintInstanceContainer();
				activateConfiguration(newObjective, constraintInstanceContainer);
			});
			String currentObjectiveName = currentDevelopmentProcess.getCurrentObjective().getName();
			currentObjectiveCombo.setText(currentObjectiveName);

			generateScoresAndUpdateUI(currentDevelopmentProcess);
			List<String> configNames = currentDevelopmentProcess.getConfigurations().stream()
					.map(c -> c.getName()).collect(toList());

			String[] configNamesArray = new String[configNames.size()];
			configNames.toArray(configNamesArray);

			currentObjectiveCombo.removeAll();

			for(String comboString : configNamesArray) {
				currentObjectiveCombo.add(comboString);
			}
			String currentObjectiveNewName =
					currentDevelopmentProcess.getCurrentObjective().getName();
			currentObjectiveCombo.setText(currentObjectiveNewName);
			// toolitem.setWidth(computeWidth(composite));
		}
	}

	// TODO (SB, 4): Also fix line-breaks, and punctuation and capitalization of @return
	/**
	 * generates an object containing how many instances of a given constraints are satisfied and
	 * how many constraints in
	 * all exist in the given objective
	 * 
	 * @param constraintName
	 * @return the value indicating whether the constraint is satisfied or not
	 */
	private ConstraintResultHolder getConstraintValueObjectByName(String constraintName) {

		Integer success = 0;
		Integer total = 0;
		EList<ConstraintInstance> instances =
				currentDevelopmentProcess.getConstraintInstanceContainer().getConstraintInstances();
		IConstraintUIService cs = IConstraintUIService.getInstance();
		for(ConstraintInstance ci : instances) {
			if(constraintName.equalsIgnoreCase(ci.getConstraintName())) {
				IConstraintInstanceStatus status = cs.getStatus(ci);
				if(status instanceof SuccessConstraintInstanceStatus) {
					success++;
				}
				total++;
			}
		}

		return new ConstraintResultHolder(success, total);
	}

	/**
	 * Generates scores for configuration.
	 * 
	 * @param process
	 *            the root process
	 * @return The class containing the total and satisfied constraint count.
	 */
	private ConstraintResultHolder
			generateScoresForConfig(ConstraintBasedDevelopmentProcess process) {

		float totalConstraintCount = 0;
		float satisfiedConstraintCount = 0;

		EList<String> activeConstraints =
				currentDevelopmentProcess.getCurrentObjective().getActiveConstraints();

		for(String activeConstraintName : activeConstraints) {
			ConstraintResultHolder constraintValueObjectByName =
					getConstraintValueObjectByName(activeConstraintName);
			satisfiedConstraintCount += constraintValueObjectByName.successCount;
			totalConstraintCount += constraintValueObjectByName.totalCount;
		}
		return new ConstraintResultHolder(satisfiedConstraintCount, totalConstraintCount);
	}

	/** {@inheritDoc} */
	@Override
	public void resourceChanged(IResourceChangeEvent event) {
		// TODO (SB, 9): Why do you check for the existence of a Display that is never used?
		if(event.getType() == IResourceChangeEvent.POST_CHANGE &&
				currentDevelopmentProcess != null && Display.getCurrent() != null) {
			generateScoresAndUpdateUI(currentDevelopmentProcess);
			EList<ConstraintConfiguration> configurations =
					currentDevelopmentProcess.getConfigurations();
			List<String> configNames =
					configurations.stream().map(c -> c.getName()).collect(toList());
			String[] configNamesArray = new String[configNames.size()];
			configNames.toArray(configNamesArray);
			currentObjectiveCombo.removeAll();

			for(String comboString : configNamesArray) {
				currentObjectiveCombo.add(comboString);
			}
			currentObjectiveCombo
					.setText(currentDevelopmentProcess.getCurrentObjective().getName());
		}
	}
}
